@@ -1,16 +1,16 @@
 from __future__ import annotations
-import sys
+import platform
 import socket
 import struct
 import threading
 from openrgb import utils
 from typing import Callable
 
-OPENRGB_PROTOCOL_VERSION = 2
+OPENRGB_PROTOCOL_VERSION = 3
 
-if sys.platform.startswith("linux"):
+if platform.system() == "Linux":
     NOSIGNAL = socket.MSG_NOSIGNAL
-elif sys.platform.startswith("win"):
+else:
     NOSIGNAL = 0
 
 
@@ -44,23 +44,19 @@ class NetworkClient:
     def start_connection(self):
         '''
         Initializes a socket, connects to the SDK, and sets the client name
-
-        :param address: the ip address of the SDK server
-        :param port: the port of the SDK server
-        :param name: the string that will be displayed on the OpenRGB SDK tab's list of clients
         '''
         if self.sock is not None:
             return
         self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        self.sock.settimeout(1.0)
 
         try:
             self.sock.connect((self.address, self.port))
-        except ConnectionRefusedError:
+        except OSError:
             self.sock = None
             raise
 
         # Checking server protocol version
-        self.sock.settimeout(1.0)
         self.send_header(0, utils.PacketType.REQUEST_PROTOCOL_VERSION, struct.calcsize('I'))
         self.send_data(struct.pack("I", self._protocol_version), False)
         try:
@@ -68,7 +64,7 @@ class NetworkClient:
         except socket.timeout:
             self._protocol_version = 0
             self.lock.release()
-        self.sock.settimeout(None)
+        self.sock.settimeout(10.0)
         # Sending the client name
         name = bytes(f"{self.name}\0", 'utf-8')
         self.send_header(0, utils.PacketType.SET_CLIENT_NAME, len(name))
@@ -78,6 +74,11 @@ class NetworkClient:
         '''
         Closes the active socket
         '''
+        try:
+            self.lock.release()
+        except Exception:
+            pass
+
         if self.sock is not None:
             self.sock.close()
             self.sock = None
@@ -86,7 +87,7 @@ class NetworkClient:
         '''
         Reads responses from the SDK
 
-        :raises OpenRGBDisconnected: when it loses connection to the SDK
+        :raises utils.OpenRGBDisconnected: when it loses connection to the SDK
         '''
         if self.sock is None:
             raise utils.OpenRGBDisconnected()
@@ -102,10 +103,12 @@ class NetworkClient:
             raise utils.OpenRGBDisconnected()
         # Unpacking the contents of the raw header struct into a list
         buff = list(struct.unpack('ccccIII', header))
-        # print(buff[:4])
+
+        # Checking packet signature
         if buff[:4] == [b'O', b'R', b'G', b'B']:
             device_id, packet_type, packet_size = buff[4:]
-            # print(device_id, packet_type, packet_size)
+
+            # Parse information from each packet type
             if packet_type == utils.PacketType.REQUEST_CONTROLLER_COUNT:
                 try:
                     buff = struct.unpack("I", self.sock.recv(packet_size))
@@ -119,9 +122,10 @@ class NetworkClient:
                         self.lock.release()
                     except RuntimeError:
                         pass
+
             elif packet_type == utils.PacketType.REQUEST_CONTROLLER_DATA:
                 try:
-                    data =  bytearray()
+                    data = bytes()
                     while len(data) < packet_size:
                         data += self.sock.recv(packet_size - len(data))
                 except utils.CONNECTION_ERRORS as e:
@@ -129,11 +133,16 @@ class NetworkClient:
                     raise utils.OpenRGBDisconnected() from e
                 finally:
                     self.lock.release()
-                self.callback(device_id, packet_type, utils.ControllerData.unpack(data, self._protocol_version))
+                try:
+                    self.callback(device_id, packet_type, utils.ControllerData.unpack(data, self._protocol_version))
+                except utils.PARSING_ERRORS as e:
+                    raise utils.ControllerParsingError(f"Unable to parse data from request `{packet_type}` for device #{device_id}") from e
+
             elif packet_type == utils.PacketType.DEVICE_LIST_UPDATED:
                 assert device_id == 0 and packet_size == 0
                 self.read()
                 self.callback(device_id, packet_type, 0)
+
             elif packet_type == utils.PacketType.REQUEST_PROTOCOL_VERSION:
                 try:
                     self.max_protocol_version = min(struct.unpack("I", self.sock.recv(packet_size))[0], OPENRGB_PROTOCOL_VERSION)
@@ -143,17 +152,21 @@ class NetworkClient:
                     raise utils.OpenRGBDisconnected() from e
                 finally:
                     self.lock.release()
+
             elif packet_type == utils.PacketType.REQUEST_PROFILE_LIST:
                 try:
-                    data =  bytearray()
+                    data = bytes()
                     while len(data) < packet_size:
                         data += self.sock.recv(packet_size - len(data))
+                    idata = iter(data)
+                    for _ in range(4):
+                        next(idata)
                 except utils.CONNECTION_ERRORS as e:
                     self.stop_connection()
                     raise utils.OpenRGBDisconnected() from e
                 finally:
                     self.lock.release()
-                self.callback(device_id, packet_type, utils.parse_list(utils.Profile, data, self._protocol_version, 4)[1])
+                self.callback(device_id, packet_type, utils.parse_list(utils.Profile, idata, self._protocol_version))
 
     def requestDeviceData(self, device: int):
         '''
@@ -175,6 +188,9 @@ class NetworkClient:
         self.read()
 
     def requestProfileList(self):
+        '''
+        Sends the request for the available profiles
+        '''
         self.send_header(0, utils.PacketType.REQUEST_PROFILE_LIST, 0)
         self.read()
 
@@ -186,16 +202,24 @@ class NetworkClient:
         :param packet_type: A utils.PacketType
         :param packet_size: The full size of the data to be sent after the header
         '''
+        self.check_version(packet_type)
         if self.sock is None:
             raise utils.OpenRGBDisconnected()
-        if packet_size > 0 or packet_type in (utils.PacketType.REQUEST_CONTROLLER_COUNT,\
-                                              utils.PacketType.REQUEST_CONTROLLER_DATA,\
-                                              utils.PacketType.REQUEST_PROTOCOL_VERSION,\
-                                              utils.PacketType.REQUEST_PROFILE_LIST):
-            self.lock.acquire()
+
+        if not self.lock.acquire(timeout=10):
+            raise utils.OpenRGBDisconnected("SDK server did not respond to previous request")
 
         try:
-            self.sock.send(struct.pack('ccccIII', b'O', b'R', b'G', b'B', device_id, packet_type, packet_size), NOSIGNAL)
+            data = struct.pack('ccccIII', b'O', b'R', b'G', b'B', device_id, packet_type, packet_size)
+            sent = self.sock.send(data, NOSIGNAL)
+            if sent != len(data):
+                self.stop_connection()
+                raise utils.OpenRGBDisconnected()
+            if packet_size == 0 and packet_type not in (utils.PacketType.REQUEST_CONTROLLER_COUNT,
+                                                        utils.PacketType.REQUEST_CONTROLLER_DATA,
+                                                        utils.PacketType.REQUEST_PROTOCOL_VERSION,
+                                                        utils.PacketType.REQUEST_PROFILE_LIST):
+                self.lock.release()
         except utils.CONNECTION_ERRORS as e:
             self.stop_connection()
             raise utils.OpenRGBDisconnected() from e
@@ -209,10 +233,27 @@ class NetworkClient:
         if self.sock is None:
             raise utils.OpenRGBDisconnected()
         try:
-            self.sock.send(data, NOSIGNAL)
+            sent = self.sock.send(data, NOSIGNAL)
+            if sent != len(data):
+                self.stop_connection()
+                raise utils.OpenRGBDisconnected()
+            if release_lock:
+                self.lock.release()
         except utils.CONNECTION_ERRORS as e:
             self.stop_connection()
             raise utils.OpenRGBDisconnected() from e
-        finally:
-            if release_lock:
-                self.lock.release()
+
+    def check_version(self, packet_type: utils.PacketType):
+        '''
+        Verifies that the packet type is supported on the version we are using.
+
+        :param packet_type: What kind of packet is going to be sent
+        :raises utils.SDKVersionError: When a packet is unsupported
+        '''
+        if self._protocol_version < 2 and packet_type in (utils.PacketType.REQUEST_PROFILE_LIST,
+                                                          utils.PacketType.REQUEST_SAVE_PROFILE,
+                                                          utils.PacketType.REQUEST_LOAD_PROFILE,
+                                                          utils.PacketType.REQUEST_DELETE_PROFILE):
+            raise utils.SDKVersionError("Profile controls not supported on protoocl versions < 2.  You probably need to update OpenRGB")
+        elif self._protocol_version < 3 and packet_type == utils.PacketType.RGBCONTROLLER_SAVEMODE:
+            raise utils.SDKVersionError("Saving modes not supported on protoocl versions < 3.  You probably need to update OpenRGB")
