@@ -1,5 +1,6 @@
 from __future__ import annotations
 import struct
+import platform
 from openrgb import utils
 from typing import Union, Optional
 from openrgb.network import NetworkClient
@@ -212,7 +213,7 @@ class Device(utils.RGBContainer):
         if not fast:
             self.update()
 
-    def set_mode(self, mode: Union[int, str, utils.ModeData]):
+    def set_mode(self, mode: Union[int, str, utils.ModeData], save: bool = False):
         '''
         Sets the device's mode
 
@@ -234,6 +235,13 @@ class Device(utils.RGBContainer):
             len(data)
         )
         self.comms.send_data(data)
+        if save:
+            self.comms.send_header(
+                self.id,
+                utils.PacketType.RGBCONTROLLER_SAVEMODE,
+                len(data)
+            )
+            self.comms.send_data(data)
         self.update()
 
     def set_custom_mode(self):
@@ -243,6 +251,18 @@ class Device(utils.RGBContainer):
             0
         )
 
+    def save_mode(self):
+        '''
+        Saves the currently selected mode
+        '''
+        data = self.modes[self.active_mode].pack(self.comms._protocol_version)
+        self.comms.send_header(
+            self.id,
+            utils.PacketType.RGBCONTROLLER_SAVEMODE,
+            len(data)
+        )
+        self.comms.send_data(data)
+
 
 class OpenRGBClient(utils.RGBObject):
     '''
@@ -329,12 +349,17 @@ class OpenRGBClient(utils.RGBObject):
 
         :param name: Can be a profile's name, index, or even the Profile itself
         :param local: Whether to load a local file or a profile from the server.
-        :param directory: what directory the profile is in.  Defaults to HOME/.config/OpenRGB
+        :param directory: what directory the profile is in.  Defaults to OpenRGB's config directory for supported OS's (Windows or Linux), or falls back to using the current working directory.
         '''
         if local:
             assert type(name) is str
             if directory == '':
-                directory = environ['HOME'].rstrip("/") + "/.config/OpenRGB"
+                if platform.system() == "Linux":
+                    directory = environ['HOME'].rstrip("/") + "/.config/OpenRGB"
+                elif platform.system() == "Windows":
+                    directory = environ['APPDATA'].rstrip("\\") + "\\OpenRGB"
+                else:
+                    directory = '.'
             with open(f'{directory}/{name}.orp', 'rb') as f:
                 controllers = utils.LocalProfile.unpack(f).controllers
                 pairs = []
@@ -354,8 +379,6 @@ class OpenRGBClient(utils.RGBObject):
                     if new_controller.active_mode != device.active_mode:
                         device.set_mode(new_controller.active_mode)
         else:
-            if self.comms._protocol_version < 2:
-                raise utils.SDKVersionError("Your version of OpenRGB doesn't support SDK profile controls")
             if type(name) is str:
                 try:
                     name = next(p for p in self.profiles if p.name.lower() == name.lower())
@@ -376,17 +399,20 @@ class OpenRGBClient(utils.RGBObject):
 
         :param name: Can be a profile's name, index, or even the Profile itself
         :param local: Whether to load a local file or a profile on the server.
-        :param directory: what directory to save the profile in.  Defaults to HOME/.config/OpenRGB
+        :param directory: what directory to save the profile in.  Defaults to OpenRGB's config directory for supported OS's (Windows or Linux), or falls back to using the current working directory.
         '''
         if local:
             self.update()
             if directory == '':
-                directory = environ['HOME'].rstrip("/") + "/.config/OpenRGB"
+                if platform.system() == "Linux":
+                    directory = environ['HOME'].rstrip("/") + "/.config/OpenRGB"
+                elif platform.system() == "Windows":
+                    directory = environ['APPDATA'].rstrip("\\") + "\\OpenRGB"
+                else:
+                    directory = '.'
             with open(f'{directory.rstrip("/")}/{name}.orp', 'wb') as f:
                 f.write(utils.Profile([dev.data for dev in self.devices]).pack())
         else:
-            if self.comms._protocol_version < 2:
-                raise utils.SDKVersionError("Your version of OpenRGB doesn't support SDK profile controls")
             if type(name) is str:
                 try:
                     name = next(p for p in self.profiles if p.name.lower() == name.lower())
@@ -407,8 +433,6 @@ class OpenRGBClient(utils.RGBObject):
 
         :param name: Can be a profile's name, index, or even the Profile itself
         '''
-        if self.comms._protocol_version < 2:
-            raise utils.SDKVersionError("Your version of OpenRGB doesn't support SDK profile controls")
         if type(name) is str:
             try:
                 name = next(p for p in self.profiles if p.name.lower() == name.lower())
@@ -437,8 +461,6 @@ class OpenRGBClient(utils.RGBObject):
         '''
         Gets the list of available profiles from the server.
         '''
-        if self.comms._protocol_version < 2:
-            raise utils.SDKVersionError("Your version of OpenRGB doesn't support SDK profile controls")
         self.comms.requestProfileList()
 
     def show(self, fast: bool = False, force: bool = False):
